//
// TempTrack_Node_ABP
//
//  Demoprogramm, das einen TTN-Node alle TX_INTERVAL Sekunden die Messwerte für Temperatur
//  und relative Luftfeuchte senden lässt, die mittels eines Wettersensors (DHT11) ermittelt
//  werden. Zusätzlich wird per GPS die aktuelle Position bestimmt und ebenfalls übertragen.
//
//  Die Session Keys werden per ABP gesetzt, ein Join wird nicht durchgeführt.
//
//  Da kein Rückkanal benötigt wird, kann hier auch ein Single-Channel-Gateway eingesetzt
//  werden. In diesem Fall ist "#define SINGLE_CHANNEL" im Kopfteil des Programms einzufügen,
//  damit die angepassten Einstellungen in der Funktion setup() aktiviert werden.
//
//  Mit dem vorliegenden Programm gelangt der Arduino Uno an seine Grenze, bzw. darüber hinaus,
//  was die Speichernutzung angeht. Durch das #define SAVESPACE werden nicht zwingend benötigte
//  Codeabschnitte nicht ins Programm aufgenommen, um Platz zu sparen.
//
//  Boardeinstellung: Arduino Uno
//  Hardware: Arduino Uno mit Dragino LoRa+GPS Shield ( <-> Definition in lmic_pins)
//  Der DHT11 ist folgendermaßen anzuschließen:
//
//  DHT-Modul         Arduino/LoRa-Shield
//    VCC---------------3V3
//    DATA--------------A1  (vgl. DHT_PIN unten)
//    GND---------------GND
//
//  Hinweis: Der DHT11 liefert nur ganzzahlige, positive Messwerte, anders als z.B. der DHT22!
//  Aus Kompatibilitätsgründen wird dennoch eine Codierung verwendet, die auch negative
//  Temperaturen sowie eine Nachkommastelle erlaubt.
//
//  Für die Jumper auf dem GPS-Shield gibt es zwei Möglichkeiten:
//
//  [A]: ohne Drähte, aber mit Eingriff beim Hochladen
//  => In diesem Fall ist #define GPS_JUMPERED zu aktivieren.
//  Hier werden die beiden GPS-Jumper (GPS_RX und GPS_TX, unterhalb der oberen Pinleiste bei
//  Pin 11 bis 9) als [O##] gesetzt, d.h. der mittlere und der rechte Pin sind verbunden.
//  Bevor Code in den Arduino hochgeladen wird, muss aber der untere Jumper entfernt werden.
//  Nach erfolgtem Upload wird der Jumper wieder gesetzt. Die Debugausgaben über die serielle 
//  Schnittstelle (USB) funktionieren trotzdem.
//
//  [B]: mit zusätzlichen Kabelverbindungen
//  => In diesem Fall muss die Zeile #define GPS_JUMPERED mit einem  vorangestellten 
//  Kommentarzeichen // auskommentiert werden, also    //#define GPS_JUMPERED
//  Hier werden die beiden Jumper entfernt (oder nur mit einer Hälfte auf den rechten Pin
//  gesteckt, um nicht verloren zu gehen, also [OO#]#. Stattdessen wird der mittlere Pin
//  der oberen Reihe (GPS_RX) über ein Dupont-Kabel (w-m) mit Pin 4 der oberen Pinleiste
//  verbunden, der mittlere Pin der unteren Reihe (GPS_TX) mit Pin 3 der oberen Pinleiste.
//  Diese Verbindungen können dauerhaft bestehen bleiben und müssen beim Upload nicht
//  entfernt werden.
//
//  Achtung: Das Programm erwartet die Datei node.h als Includedatei im selben Verzeichnis.
//  Darin sind die LoRaWAN-Parameter DEVADDR, APPSKEY und NWKSKEY sowie der Devicename DEVICE
//  als Konstantendefinitionen enthalten.
//  Diese Datei muss erstellt werden bzw. wird zur Verfügung gestellt.
//  Beispielhafter Inhalt der Datei:
//
//        static const char* DEVICE = "device-abp-00";
//        static const u1_t DEVEUI[8] = { 0x77, 0x66, ......, 0x00 };   // nicht verwendet
//        static const u1_t APPEUI[8] = { 0x77, 0x66, ......, 0x00 };   // nicht verwendet
//        static const u4_t DEVADDR = 0x2601....;
//        static const u1_t APPSKEY[16] = { 0x00, 0x11, ......, 0xFF };  // 16 bytes, MSB
//        static const u1_t NWKSKEY[16] = { 0x00, 0x11, ......, 0xFF };  // 16 bytes, MSB
//
#include <lmic.h>
#include <hal/hal.h>

#include "node.h"

// bei Bedarf aktivieren:
//#define GPS_JUMPERED

// bei Bedarf aktivieren:
//#define SINGLE_CHANNEL

// bei Bedarf deaktivieren:
#define SAVESPACE

#include <dht.h>
#define DHT_PIN A0

dht DHT;

#include <SoftwareSerial.h>
#include <TinyGPS.h>

#ifdef GPS_JUMPERED
SoftwareSerial ss(0, 1); // GPS RX = Arduino D0, GPS TX = Arduino D1 (via jumpers)
#else
SoftwareSerial ss(4, 3); // GPS RX = Arduino D4, GPS TX = Arduino D3 (via cables)
#endif
TinyGPS gps;

unsigned int seqno = 0;
static uint8_t mydata[18];

// These callbacks are only used in over-the-air activation, so they are
// left empty here (we cannot leave them out completely unless
// DISABLE_JOIN is set in config.h, otherwise the linker will complain).
void os_getArtEui (u1_t* buf) { }
void os_getDevEui (u1_t* buf) { }
void os_getDevKey (u1_t* buf) { }

static osjob_t sendjob;

// Schedule TX every this many seconds (might become longer due to duty
// cycle limitations).
const unsigned TX_INTERVAL = 10;
const int LORAWAN_PORT = 3;

// Pin mapping
const lmic_pinmap lmic_pins = {
  .nss = 10,
  .rxtx = LMIC_UNUSED_PIN,
  .rst = 9,
  .dio = {2, 6, 7},
};

void onEvent (ev_t ev) {
  Serial.print(os_getTime());
  Serial.print(": ");
  switch (ev) {
#ifndef SAVESPACE
    case EV_SCAN_TIMEOUT:
      Serial.println(F("EV_SCAN_TIMEOUT"));
      break;
    case EV_BEACON_FOUND:
      Serial.println(F("EV_BEACON_FOUND"));
      break;
    case EV_BEACON_MISSED:
      Serial.println(F("EV_BEACON_MISSED"));
      break;
    case EV_BEACON_TRACKED:
      Serial.println(F("EV_BEACON_TRACKED"));
      break;
    case EV_JOINING:
      Serial.println(F("EV_JOINING"));
      break;
    case EV_JOINED:
      Serial.println(F("EV_JOINED"));
      break;
    case EV_RFU1:
      Serial.println(F("EV_RFU1"));
      break;
    case EV_JOIN_FAILED:
      Serial.println(F("EV_JOIN_FAILED"));
      break;
    case EV_REJOIN_FAILED:
      Serial.println(F("EV_REJOIN_FAILED"));
      break;
#endif
    case EV_TXCOMPLETE:
      Serial.print(F("EV_TXCOMPLETE (includes waiting for RX windows); frequency = "));
      Serial.println(LMIC.freq);
      if (LMIC.dataLen) {
        // data received in rx slot after tx
        Serial.print(F("Data Received: "));
        Serial.write(LMIC.frame + LMIC.dataBeg, LMIC.dataLen);
        Serial.println();
      }
      // Schedule next transmission
      os_setTimedCallback(&sendjob, os_getTime() + sec2osticks(TX_INTERVAL), do_send);
      break;
#ifndef SAVESPACE
    case EV_LOST_TSYNC:
      Serial.println(F("EV_LOST_TSYNC"));
      break;
    case EV_RESET:
      Serial.println(F("EV_RESET"));
      break;
    case EV_RXCOMPLETE:
      // data received in ping slot
      Serial.println(F("EV_RXCOMPLETE"));
      break;
    case EV_LINK_DEAD:
      Serial.println(F("EV_LINK_DEAD"));
      break;
    case EV_LINK_ALIVE:
      Serial.println(F("EV_LINK_ALIVE"));
      break;
#ifdef _LMIC_UTIL_H_
    case EV_SCAN_FOUND:
      Serial.println(F("EV_SCAN_FOUND"));
      break;
    case EV_TXSTART:
      Serial.println(F("EV_TXSTART"));
      break;
#endif
#endif
    default:
      Serial.println(F("Unknown event"));
      break;
  }
}

void do_send(osjob_t* j) {
  // Check if there is not a current TX/RX job running
  if (LMIC.opmode & OP_TXRXPEND) {
    Serial.println(F("OP_TXRXPEND, not sending"));
  } else {

#ifdef SAVESPACE
    Serial.println(String(F("###########    ")) + F("No.") + seqno + F("    ###########"));
#endif

    DHT.read11(DHT_PIN);    //Temperature detection
    int temp = DHT.temperature;
    int rhum = DHT.humidity;
#ifndef SAVESPACE
    Serial.println(F("Temperature and humidity :"));
    Serial.println(String(F(" raw                [")) + temp + F("°, ") + rhum + F("%]"));
#endif

#define MSB(val) (uint8_t(val >> 8))
#define LSB(val) (uint8_t(val & 0xff))

    mydata[0] = MSB(seqno);
    mydata[1] = LSB(seqno);
    int16_t temp_i = (temp * 10);
    mydata[2] = MSB(temp_i);
    mydata[3] = LSB(temp_i);
    int16_t rhum_i = (rhum * 10);
    mydata[4] = MSB(rhum_i);
    mydata[5] = LSB(rhum_i);
#ifdef SAVESPACE
    Serial.println(String(F(" transport encoding [")) + temp_i + F(", ") + rhum_i + F("]"));
#endif

    float fhdop, flat, flon, falt;
    unsigned long age;
    fhdop = gps.hdop();
    gps.f_get_position(&flat, &flon, &age);
    falt = gps.f_altitude(); //get altitude
/*
    flon == TinyGPS::GPS_INVALID_F_ANGLE ? 0.0 : flon, 6; //save six decimal places
    flat == TinyGPS::GPS_INVALID_F_ANGLE ? 0.0 : flat, 6;
    falt == TinyGPS::GPS_INVALID_F_ANGLE ? 0.0 : falt, 2; //save two decimal places
*/

    long hdop, lat, lon, alt;
    hdop = fhdop;
    lat = floor(flat * 1000000.0);
    lon = floor(flon * 1000000.0);
    alt = floor(falt);

#ifndef SAVESPACE
    Serial.println(String(F("GPS readout: [HDOP "))
                   + fhdop + (", lat ") + flat + F("°, lon ") + flon + F("°, alt ") + falt + F(" m]\n"));
#else
    Serial.println(String(F("GPS readout: [HDOP "))
                   + hdop + (", lat ") + lat + F(" µ°, lon ") + lon + F(" µ°, alt ") + alt + F(" cm]\n"));
#endif
    
    mydata[6] =  (hdop & 0xff00) >> 8;
    mydata[7] =  (hdop & 0x00ff);
    mydata[8] =  (lat & 0xff000000) >> 24;
    mydata[9] =  (lat & 0x00ff0000) >> 16;
    mydata[10] =  (lat & 0x0000ff00) >> 8;
    mydata[11] =  (lat & 0x000000ff);
    mydata[12] =  (lon & 0xff000000) >> 24;
    mydata[13] =  (lon & 0x00ff0000) >> 16;
    mydata[14] =  (lon & 0x0000ff00) >> 8;
    mydata[15] =  (lon & 0x000000ff);
    mydata[16] =  (alt & 0x0000ff00) >> 8;
    mydata[17] =  (alt & 0x000000ff);

#ifdef SAVESPACE
    Serial.print(F(" telegram bytes     ["));
    for (int i = 0; i < sizeof(mydata); i++) {
      Serial.print(mydata[i]);
      Serial.print(" ");
    }
    Serial.println("]");
#endif

    // Prepare upstream data transmission at the next possible time.
    LMIC_setTxData2(LORAWAN_PORT, mydata, sizeof(mydata), 0);
    Serial.println(F("Packet queued"));

    seqno++;
  }
  // Next TX is scheduled after TX_COMPLETE event.
}

void setup() {
  Serial.begin(115200);
  Serial.println(String(F("\n\n Starting TempTrack_Node_ABP for device ")) + DEVICE);

  // GPS interface init
  ss.begin(9600);

  // LMIC init
  os_init();
  // Reset the MAC state. Session and pending data transfers will be discarded.
  LMIC_reset();

  // Set static session parameters. Instead of dynamically establishing a session
  // by joining the network, precomputed session parameters are be provided.
  LMIC_setSession (0x1, DEVADDR, NWKSKEY, APPSKEY);

#ifdef SINGLE_CHANNEL
  Serial.println("< Single channel mode >");
  // Set up the channels used by the Things Network, which corresponds
  // to the defaults of most gateways. Without this, only three base
  // channels from the LoRaWAN specification are used, which certainly
  // works, so it is good for debugging, but can overload those
  // frequencies, so be sure to configure the full frequency range of
  // your network here (unless your network autoconfigures them).
  // Setting up channels should happen after LMIC_setSession, as that
  // configures the minimal channel set.
  LMIC_setupChannel(0, 868100000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
  for (int i = 1; i <= 8; i++) LMIC_disableChannel(i);
  /*
    LMIC_setupChannel(1, 868300000, DR_RANGE_MAP(DR_SF12, DR_SF7B), BAND_CENTI);      // g-band
    LMIC_setupChannel(2, 868500000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
    LMIC_setupChannel(3, 867100000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
    LMIC_setupChannel(4, 867300000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
    LMIC_setupChannel(5, 867500000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
    LMIC_setupChannel(6, 867700000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
    LMIC_setupChannel(7, 867900000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
    LMIC_setupChannel(8, 868800000, DR_RANGE_MAP(DR_FSK,  DR_FSK),  BAND_MILLI);      // g2-band
  */
  // TTN defines an additional channel at 869.525Mhz using SF9 for class B
  // devices' ping slots. LMIC does not have an easy way to define set this
  // frequency and support for class B is spotty and untested, so this
  // frequency is not configured here.
#endif

  // Disable link check validation
  LMIC_setLinkCheckMode(0);

  // Set data rate and transmit power (note: txpow seems to be ignored by the library)
  LMIC_setDrTxpow(DR_SF7, 14);

  // Start job
  do_send(&sendjob);
}

static void smartdelay(unsigned long ms)
{
  unsigned long start = millis();
  do
  {
    while (ss.available())
    {
      //ss.print(Serial.read());
      gps.encode(ss.read());
    }
  } while (millis() - start < ms);
}
void loop() {
  os_runloop_once();
  smartdelay(1000);
}
